// Definition d'un registre Intel 8086
typedef union
{
  struct
  {
   byte Lo;
   byte Hi;
  } Octet;  // Un octet...
  word Mot; // ...ou un mot
} Registre8086;

// Definition d'un registre Intel 32 bits
typedef union
{
  struct
  {
   word Lo;
   word Hi;
  } Mot;  // Un mot...
  dword DMot; // ...ou un double mot
} Registre32;


// Registres generaux
Registre8086 Registre_AX;   // Accumulateur
Registre8086 Registre_DX;
Registre8086 Registre_CX;
Registre8086 Registre_BX;
Registre8086 Registre_BP;   // Base Pointer
Registre8086 Registre_SI;   // Source Index
Registre8086 Registre_DI;   // Destination Index
Registre8086 Registre_SP;   // Stack Pointer

// Registres de segments
Registre8086 Registre_CS;   // Code Segment
Registre8086 Registre_SS;   // Stack Segment
Registre8086 Registre_DS;   // Data Segment
Registre8086 Registre_ES;   // Extra Segment
/* 80386
Registre8086 Registre_FS;   // Data Segment
Registre8086 Registre_GS;   // Data Segment
*/

Registre8086 Registre_IP;   // Instruction Pointer

Registre8086 Registre_CALC;  // Registre de calcul 16 bits
Registre32   Registre_LCALC; // Registre de calcul 32 bits

static byte CF,PF,AF,ZF,SF,TF,IF,DF,OF;

// Les registres Intel 8086
#define AX    Registre_AX.Mot        // Raccourci pour AX
#define AH    Registre_AX.Octet.Hi   // -> AH
#define AL    Registre_AX.Octet.Lo   // -> AL
#define DX    Registre_DX.Mot        // Raccourci pour DX
#define DH    Registre_DX.Octet.Hi   // -> DH
#define DL    Registre_DX.Octet.Lo   // -> DL
#define CX    Registre_CX.Mot        // Raccourci pour CX
#define CH    Registre_CX.Octet.Hi   // -> CH
#define CL    Registre_CX.Octet.Lo   // -> CL
#define BX    Registre_BX.Mot        // Raccourci pour BX
#define BH    Registre_BX.Octet.Hi   // -> BH
#define BL    Registre_BX.Octet.Lo   // -> BL
#define BP    Registre_BP.Mot        // Raccourci pour BP
#define SI    Registre_SI.Mot        // Raccourci pour SI
#define DI    Registre_DI.Mot        // Raccourci pour DI
#define SP    Registre_SP.Mot        // Raccourci pour SP

#define CS    Registre_CS.Mot        // Raccourci pour CS
#define SS    Registre_SS.Mot        // Raccourci pour SS
#define DS    Registre_DS.Mot        // Raccourci pour DS
#define ES    Registre_ES.Mot        // Raccourci pour ES

#define IP    Registre_IP.Mot        // Raccourci pour IP

#define CALC  Registre_CALC.Mot      // Raccourci pour CALC (auxiliaire 16 bits)
#define CALCH Registre_CALC.Octet.Hi // -> CALC partie haute
#define CALCL Registre_CALC.Octet.Lo // -> CALC partie basse

#define LCALC  Registre_LCALC.DMot   // Raccourci pour LCALC (auxiliaire 32 bits)
#define LCALCH Registre_LCALC.Mot.Hi // -> LCALC partie haute
#define LCALCL Registre_LCALC.Mot.Lo // -> LCALC partie basse

byte *R8[8] =  {&AL,&CL,&DL,&BL,&AH,&CH,&DH,&BH };
word *R16[8] = {&AX,&CX,&DX,&BX,&SP,&BP,&SI,&DI };
word *SEG[8] = {&ES,&CS,&SS,&DS,NULL,NULL,NULL,NULL };

byte modrm,regop,mod,rm;
byte val8,tmp8;
word val16,tmp16;
dword adr;

#define GEN_CF8(R)    CF=((R&0x100)==0x100)
#define GEN_CF16(R)   CF=((R&0x10000)==0x10000)
#define GEN_AF(X,Y,R) AF=(((X^Y^R)&0x10)==0x10)
#define GEN_SF8(R)    SF=((R&0x80)==0x80)
#define GEN_SF16(R)   SF=((R&0x8000)==0x8000)
#define GEN_ZF(R)     ZF=(!R)
#define GEN_PF(R)     PF=(PAIR[R&0xFF])

#define FETCH(R)  R=FetchB()
#define FETCHW(R) R=FetchB(); R+=(FetchB()<<8)

#define READBYTE(R,A)  R=ReadMemory(A)
#define READWORD(R,A) R=ReadMemory(A); R+=ReadMemory(A+1)<<8

#define PUTWORD(ADR,VALUE) \
ActionMemory(ADR,VALUE&0xFF); \
ActionMemory(ADR+1,VALUE>>8); \

#define GEN_STOSW \
ActionMemory((ES<<4)+DI,AL); \
ActionMemory((ES<<4)+DI+1,AH); \
DI+=DF?-2:2; \
IP+=1; \

#define GEN_OUTSB \
READBYTE(tmp8,(DS<<4)+SI); \
OutPortB(DX,tmp8); \
SI+=DF?-1:1; \

#define GEN_REP \
FETCH(tmp8); \
switch(tmp8) \
{ \
/* OUTSB */ case 0x6E : for(; CX>0; CX-=1) {GEN_OUTSB;} IP+=1; break; \
}; \

#define REGOP regop=(modrm&0x38)>>3
#define RM    rm=modrm&0x07
#define MOD   mod=modrm>>6

#define DECODRM FETCH(modrm); REGOP; MOD; RM

#define GEN_PUSH(X) SP-=2;  PUTWORD((SS<<4)+SP,X);
#define GEN_POP(X)  READWORD(X,(SS<<4)+SP); SP+=2;

#define GEN_PUSH_imm8  FETCH(tmp8);   GEN_PUSH(tmp8);  IP+=1;
#define GEN_PUSH_imm16 FETCHW(tmp16); GEN_PUSH(tmp16); IP+=1;

#define GEN_PUSHA \
tmp16=SP; \
GEN_PUSH(AX); \
GEN_PUSH(CX); \
GEN_PUSH(DX); \
GEN_PUSH(BX); \
GEN_PUSH(tmp16); \
GEN_PUSH(BP); \
GEN_PUSH(SI); \
GEN_PUSH(DI); \
IP+=1; \

#define GEN_POPA \
GEN_POP(DI); \
GEN_POP(SI); \
GEN_POP(BP); \
GEN_POP(tmp16); \
GEN_POP(BX); \
GEN_POP(DX); \
GEN_POP(CX); \
GEN_POP(AX); \
IP+=1; \

#define GEN_JMP_rel8(X) FETCH(tmp8); IP+=X?(signed char)tmp8+1:1;

#define GEN_AND(R,X) R&=X; CF=OF=0; GEN_ZF(R); GEN_PF(R)
#define GEN_AND8(R,X)  GEN_AND(R,X); GEN_SF8(R)
#define GEN_AND16(R,X) GEN_AND(R,X); GEN_SF16(R)
#define GEN_AND_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_AND8(*R8[regop],val8);    IP+=1
#define GEN_AND_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_AND16(*R16[regop],val16); IP+=1
#define GEN_AND_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_AND8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_AND_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_AND16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_AND_AL_imm8  FETCH(tmp8);   GEN_AND8(AL,tmp8);   IP+=1;
#define GEN_AND_AX_imm16 FETCHW(tmp16); GEN_AND16(AX,tmp16); IP+=1;
#define GEN_AND_rm8_imm8   FETCH(tmp8);   GEN_AND8(val8,tmp8);    IP+=1
#define GEN_AND_rm16_rel8  FETCH(tmp8);   GEN_AND16(val16,(signed char)tmp8);    IP+=1
#define GEN_AND_rm16_imm16 FETCHW(tmp16); GEN_AND16(val16,tmp16); IP+=1

#define GEN_OR(R,X)  R|=X; CF=OF=0; GEN_ZF(R); GEN_PF(R)
#define GEN_OR8(R,X)  GEN_OR(R,X); GEN_SF8(R)
#define GEN_OR16(R,X) GEN_OR(R,X); GEN_SF16(R)
#define GEN_OR_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_OR8(*R8[regop],val8);    IP+=1
#define GEN_OR_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_OR16(*R16[regop],val16); IP+=1
#define GEN_OR_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_OR8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_OR_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_OR16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_OR_AL_imm8  FETCH(tmp8);   GEN_OR8(AL,tmp8);   IP+=1;
#define GEN_OR_AX_imm16 FETCHW(tmp16); GEN_OR16(AX,tmp16); IP+=1;
#define GEN_OR_rm8_imm8   FETCH(tmp8);   GEN_OR8(val8,tmp8);    IP+=1
#define GEN_OR_rm16_rel8  FETCH(tmp8);   GEN_OR16(val16,(signed char)tmp8);    IP+=1
#define GEN_OR_rm16_imm16 FETCHW(tmp16); GEN_OR16(val16,tmp16); IP+=1

#define GEN_XOR(R,X) R^=X; CF=OF=0; GEN_ZF(R); GEN_PF(R)
#define GEN_XOR8(R,X)  GEN_XOR(R,X); GEN_SF8(R)
#define GEN_XOR16(R,X) GEN_XOR(R,X); GEN_SF16(R)
#define GEN_XOR_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_XOR8(*R8[regop],val8);    IP+=1
#define GEN_XOR_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_XOR16(*R16[regop],val16); IP+=1
#define GEN_XOR_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_XOR8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_XOR_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_XOR16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_XOR_AL_imm8  FETCH(tmp8);   GEN_XOR8(AL,tmp8);   IP+=1;
#define GEN_XOR_AX_imm16 FETCHW(tmp16); GEN_XOR16(AX,tmp16); IP+=1;
#define GEN_XOR_rm8_imm8   FETCH(tmp8);   GEN_XOR8(val8,tmp8);    IP+=1
#define GEN_XOR_rm16_rel8  FETCH(tmp16);  GEN_XOR16(val16,(signed char)tmp8);    IP+=1
#define GEN_XOR_rm16_imm16 FETCHW(tmp16); GEN_XOR16(val16,tmp16); IP+=1

#define GEN_MOV_r16_imm16(R) FETCHW(R); IP+=1
#define GEN_MOV_r8_imm8(R)   FETCH(R);  IP+=1

#define GEN_MOD00(ADR) \
switch(rm) \
{ \
case 0x00 : ADR=(DS<<4)+BX+SI; break; \
case 0x01 : ADR=(DS<<4)+BX+DI; break; \
case 0x02 : ADR=(SS<<4)+BP+SI; break; \
case 0x03 : ADR=(SS<<4)+BP+DI; break; \
case 0x04 : ADR=(DS<<4)+SI; break; \
case 0x05 : ADR=(DS<<4)+DI; break; \
case 0x06 : FETCHW(ADR); ADR+=(DS<<4); break; \
case 0x07 : ADR=(DS<<4)+BX; break; \
} \         

#define GEN_MOD01(ADR) \
switch(rm) \
{ \                           
case 0x00 : FETCH(ADR); ADR+=(DS<<4)+BX+SI; break; \
case 0x01 : FETCH(ADR); ADR+=(DS<<4)+BX+DI; break; \
case 0x02 : FETCH(ADR); ADR+=(SS<<4)+BP+SI; break; \
case 0x03 : FETCH(ADR); ADR+=(SS<<4)+BP+DI; break; \
case 0x04 : FETCH(ADR); ADR+=(DS<<4)+SI; break; \
case 0x05 : FETCH(ADR); ADR+=(DS<<4)+DI; break; \
case 0x06 : FETCH(ADR); ADR+=(SS<<4)+BP; break; \
case 0x07 : FETCH(ADR); ADR+=(DS<<4)+BX; break; \
} \         

#define GEN_MOD10(ADR) \
switch(rm) \
{ \
case 0x00 : FETCHW(ADR); ADR+=(DS<<4)+BX+SI; break; \
case 0x01 : FETCHW(ADR); ADR+=(DS<<4)+BX+DI; break; \
case 0x02 : FETCHW(ADR); ADR+=(DS<<4)+BX+SI; break; \
case 0x03 : FETCHW(ADR); ADR+=(DS<<4)+BX+DI; break; \
case 0x04 : FETCHW(ADR); ADR+=(DS<<4)+SI; break; \
case 0x05 : FETCHW(ADR); ADR+=(DS<<4)+DI; break; \
case 0x06 : FETCHW(ADR); ADR+=(SS<<4)+BP; break; \
case 0x07 : FETCHW(ADR); ADR+=(DS<<4)+BX; break; \
} \         

#define GEN_GETRM8(ADR,VALUE) \
{ \
 switch(mod) \
 { \
 case 0x00 : GEN_MOD00(ADR); VALUE=ReadMemory(ADR); break; \
 case 0x01 : GEN_MOD01(ADR); VALUE=ReadMemory(ADR); break; \
 case 0x02 : GEN_MOD10(ADR); VALUE=ReadMemory(ADR); break; \
 case 0x03 : VALUE=*R8[rm]; break; \
 } \
} \

#define GEN_PUTRM8(ADR,VALUE) \
if (mod==0x03) *R8[rm]=VALUE; else ActionMemory(ADR,VALUE); \

#define GEN_GETRM16(ADR,VALUE) \
switch(mod) \
{ \
case 0x00 : GEN_MOD00(ADR); VALUE=ReadMemory(ADR)+(ReadMemory(ADR+1)<<8); break; \
case 0x01 : GEN_MOD01(ADR); VALUE=ReadMemory(ADR)+(ReadMemory(ADR+1)<<8); break; \
case 0x02 : GEN_MOD10(ADR); VALUE=ReadMemory(ADR)+(ReadMemory(ADR+1)<<8); break; \
case 0x03 : VALUE=*R16[rm]; break; \
} \

#define GEN_PUTRM16(ADR,VALUE) \
if (mod==0x03) *R16[rm]=VALUE; else PUTWORD(ADR,VALUE); \

#define GEN_MOV_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);   *R8[regop]=val8;   IP+=1
#define GEN_MOV_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); *R16[regop]=val16; IP+=1
#define GEN_MOV_seg_rm16 DECODRM; GEN_GETRM16(adr,val16); *SEG[regop]=val16; IP+=1

#define GEN_MOV_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);   GEN_PUTRM8(adr,*R8[regop]);   IP+=1
#define GEN_MOV_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_PUTRM16(adr,*R16[regop]); IP+=1
#define GEN_MOV_rm16_seg DECODRM; GEN_GETRM16(adr,val16); GEN_PUTRM16(adr,*SEG[regop]); IP+=1

#define GEN_XCHG_rm8_r8  DECODRM; GEN_GETRM8(adr,val8);   GEN_PUTRM8(adr,*R8[regop]); *R8[regop]=val8; IP+=1

#define GEN_INC8(R)  CALC=++R;  GEN_AF(CALC,R,1); GEN_SF8(CALC); GEN_ZF(CALC); GEN_PF(CALC); R=CALC; 
#define GEN_DEC8(R)  CALC=--R;  GEN_AF(CALC,R,1); GEN_SF8(CALC); GEN_ZF(CALC); GEN_PF(CALC); R=CALC; 
#define GEN_INC16(R) CALC=++R;  GEN_AF(CALC,R,1); GEN_SF16(CALC); GEN_ZF(CALC); GEN_PF(CALC); R=CALC; 
#define GEN_DEC16(R) CALC=--R;  GEN_AF(CALC,R,1); GEN_SF16(CALC); GEN_ZF(CALC); GEN_PF(CALC); R=CALC; 

#define GEN_ADD8(R,X)  CALC=R+X;  GEN_CF8(CALC);   GEN_AF(CALC,X,R);  GEN_SF8(CALC);   GEN_ZF(CALC);  GEN_PF(CALC); R=CALCL
#define GEN_ADD16(R,X) LCALC=R+X; GEN_CF16(LCALC); GEN_AF(LCALC,X,R); GEN_SF16(LCALC); GEN_ZF(LCALC); GEN_PF(LCALC); R=LCALCL
#define GEN_ADD_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_ADD8(*R8[regop],val8);    IP+=1
#define GEN_ADD_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_ADD16(*R16[regop],val16); IP+=1
#define GEN_ADD_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_ADD8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_ADD_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_ADD16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_ADD_AL_imm8  FETCH(tmp8);   GEN_ADD8(AL,tmp8);   IP+=1;
#define GEN_ADD_AX_imm16 FETCHW(tmp16); GEN_ADD16(AX,tmp16); IP+=1;
#define GEN_ADD_rm8_imm8   FETCH(tmp8);   GEN_ADD8(val8,tmp8);    IP+=1
#define GEN_ADD_rm16_rel8  FETCH(tmp8);   GEN_ADD16(val16,(signed char)tmp8);    IP+=1
#define GEN_ADD_rm16_imm16 FETCHW(tmp16); GEN_ADD16(val16,tmp16); IP+=1

#define GEN_ADC8(R,X)  CALC=R+X+CF;  GEN_CF8(CALC);   GEN_AF(CALC,X,R);  GEN_SF8(CALC);   GEN_ZF(CALC);  GEN_PF(CALC); R=CALCL
#define GEN_ADC16(R,X) LCALC=R+X+CF; GEN_CF16(LCALC); GEN_AF(LCALC,X,R); GEN_SF16(LCALC); GEN_ZF(LCALC); GEN_PF(LCALC); R=LCALCL
#define GEN_ADC_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_ADC8(*R8[regop],val8);    IP+=1
#define GEN_ADC_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_ADC16(*R16[regop],val16); IP+=1
#define GEN_ADC_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_ADC8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_ADC_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_ADC16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_ADC_AL_imm8  FETCH(tmp8);   GEN_ADC8(AL,tmp8);   IP+=1;
#define GEN_ADC_AX_imm16 FETCHW(tmp16); GEN_ADC16(AX,tmp16); IP+=1;
#define GEN_ADC_rm8_imm8   FETCH(tmp8);   GEN_ADC8(val8,tmp8);    IP+=1
#define GEN_ADC_rm16_rel8  FETCH(tmp8);   GEN_ADC16(val16,(signed char)tmp8);    IP+=1
#define GEN_ADC_rm16_imm16 FETCHW(tmp16); GEN_ADC16(val16,tmp16); IP+=1

#define GEN_NEG8(X)  CALC=-X;  GEN_CF8(CALC);   GEN_AF(CALC,X,0);  GEN_SF8(CALC);   GEN_ZF(CALC);  GEN_PF(CALC); X=CALCL
#define GEN_NEG16(X) LCALC=-X; GEN_CF16(LCALC); GEN_AF(LCALC,X,0); GEN_SF16(LCALC); GEN_ZF(LCALC); GEN_PF(LCALC); X=LCALCL

#define GEN_SUB8(R,X)  CALC=R-X;  GEN_CF8(CALC);   GEN_AF(CALC,X,R);  GEN_SF8(CALC);   GEN_ZF(CALC);  GEN_PF(CALC); R=CALCL
#define GEN_SUB16(R,X) LCALC=R-X; GEN_CF16(LCALC); GEN_AF(LCALC,X,R); GEN_SF16(LCALC); GEN_ZF(LCALC); GEN_PF(LCALC); R=LCALCL
#define GEN_SUB_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_SUB8(*R8[regop],val8);    IP+=1
#define GEN_SUB_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_SUB16(*R16[regop],val16); IP+=1
#define GEN_SUB_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_SUB8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_SUB_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_SUB16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_SUB_AL_imm8  FETCH(tmp8);   GEN_SUB8(AL,tmp8);   IP+=1;
#define GEN_SUB_AX_imm16 FETCHW(tmp16); GEN_SUB16(AX,tmp16); IP+=1;
#define GEN_SUB_rm8_imm8   FETCH(tmp8);   GEN_SUB8(val8,tmp8);    IP+=1
#define GEN_SUB_rm16_rel8  FETCH(tmp8);   GEN_SUB16(val16,(signed char)tmp8);    IP+=1
#define GEN_SUB_rm16_imm16 FETCHW(tmp16); GEN_SUB16(val16,tmp16); IP+=1

#define GEN_CMP8(R,X)  CALC=R-X;  GEN_CF8(CALC);   GEN_AF(CALC,X,R);  GEN_SF8(CALC);   GEN_ZF(CALC);  GEN_PF(CALC)
#define GEN_CMP16(R,X) LCALC=R-X; GEN_CF16(LCALC); GEN_AF(LCALC,X,R); GEN_SF16(LCALC); GEN_ZF(LCALC); GEN_PF(LCALC)
#define GEN_CMP_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_CMP8(*R8[regop],val8);    IP+=1
#define GEN_CMP_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_CMP16(*R16[regop],val16); IP+=1
#define GEN_CMP_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_CMP8(val8,*R8[regop]);    IP+=1
#define GEN_CMP_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_CMP16(val16,*R16[regop]); IP+=1
#define GEN_CMP_AL_imm8  FETCH(tmp8);   GEN_CMP8(AL,tmp8);   IP+=1;
#define GEN_CMP_AX_imm16 FETCHW(tmp16); GEN_CMP16(AX,tmp16); IP+=1;
#define GEN_CMP_rm8_imm8   FETCH(tmp8);   GEN_CMP8(val8,tmp8);    IP+=1
#define GEN_CMP_rm16_rel8  FETCH(tmp8);   GEN_CMP16(val16,(signed char)tmp8);    IP+=1
#define GEN_CMP_rm16_imm16 FETCHW(tmp16); GEN_CMP16(val16,tmp16); IP+=1

#define GEN_SBB8(R,X)  CALC=R-X-CF;  GEN_CF8(CALC);   GEN_AF(CALC,X,R);  GEN_SF8(CALC);   GEN_ZF(CALC);  GEN_PF(CALC); R=CALCL
#define GEN_SBB16(R,X) LCALC=R-X-CF; GEN_CF16(LCALC); GEN_AF(LCALC,X,R); GEN_SF16(LCALC); GEN_ZF(LCALC); GEN_PF(LCALC); R=LCALCL
#define GEN_SBB_r8_rm8   DECODRM; GEN_GETRM8(adr,val8);    GEN_SBB8(*R8[regop],val8);    IP+=1
#define GEN_SBB_r16_rm16 DECODRM; GEN_GETRM16(adr,val16); GEN_SBB16(*R16[regop],val16); IP+=1
#define GEN_SBB_rm8_r8   DECODRM; GEN_GETRM8(adr,val8);    GEN_SBB8(val8,*R8[regop]);    GEN_PUTRM8(adr,val8);     IP+=1
#define GEN_SBB_rm16_r16 DECODRM; GEN_GETRM16(adr,val16); GEN_SBB16(val16,*R16[regop]); GEN_PUTRM16(adr,val16);  IP+=1
#define GEN_SBB_AL_imm8  FETCH(tmp8);   GEN_SBB8(AL,tmp8);   IP+=1;
#define GEN_SBB_AX_imm16 FETCHW(tmp16); GEN_SBB16(AX,tmp16); IP+=1;
#define GEN_SBB_rm8_imm8   FETCH(tmp8);   GEN_SBB8(val8,tmp8);    IP+=1
#define GEN_SBB_rm16_rel8  FETCH(tmp8);   GEN_SBB16(val16,(signed char)tmp8);    IP+=1
#define GEN_SBB_rm16_imm16 FETCHW(tmp16); GEN_SBB16(val16,tmp16); IP+=1

#define GEN_ROL_rm8_1(RM) IP+=1
#define GEN_ROR_rm8_1(RM) IP+=1
#define GEN_RCL_rm8_1(RM) IP+=1
#define GEN_RCR_rm8_1(RM) IP+=1
#define GEN_SHL_rm8_1(RM) *R8[RM]<<=1; IP+=1
#define GEN_SHR_rm8_1(RM) *R8[RM]>>=1; IP+=1
#define GEN_SAR_rm8_1(RM) *R8[RM]>>=1; IP+=1

#define GEN_MUL_rm8(VALUE)  AX*=*R8[RM]; IP+=1
#define GEN_IMUL_rm8(VALUE) AX*=*R8[RM]; IP+=1
#define GEN_DIV_rm8(VALUE)  AX/=*R8[RM]; IP+=1
#define GEN_IDIV_rm8(VALUE) AX/=*R8[RM]; IP+=1

#define GEN_TEST_rm8_imm8(VALUE) FETCH(tmp8); GEN_AND8(tmp8,VALUE); IP+=1;

#define GEN_MUL_rm16(VALUE) \
LCALC=AX*VALUE; \
AX=LCALCL; \
DX=LCALCH; \
CF=OF=(DX!=0); \
IP+=1; \

#define GEN_IMUL_rm16(VALUE) AX*=VALUE; IP+=1

#define GEN_DIV_rm16(VALUE) \
LCALC=(DX<<16)+AX; \
DX=LCALC%VALUE; \
AX=LCALC/VALUE; \
IP+=1 \

#define GEN_IDIV_rm16(VALUE) AX/=*R16[RM]; IP+=1

#define GEN_TEST_rm16_imm16(VALUE) FETCHW(tmp16); GEN_AND16(tmp16,VALUE); IP+=1;

#define GEN_NOT(VALUE)  VALUE=~VALUE; IP+=1

#define GEN_GR1_rm8_imm8 \
DECODRM; \
GEN_GETRM8(adr,val8); \
switch(regop) \
{ \
case 0x00 : GEN_ADD_rm8_imm8; GEN_PUTRM8(adr,val8); break; \
case 0x01 : GEN_OR_rm8_imm8;  GEN_PUTRM8(adr,val8); break; \
case 0x02 : GEN_ADC_rm8_imm8; GEN_PUTRM8(adr,val8); break; \
case 0x03 : GEN_SBB_rm8_imm8; GEN_PUTRM8(adr,val8); break; \
case 0x04 : GEN_AND_rm8_imm8; GEN_PUTRM8(adr,val8); break; \
case 0x05 : GEN_SUB_rm8_imm8; GEN_PUTRM8(adr,val8); break; \
case 0x06 : GEN_XOR_rm8_imm8; GEN_PUTRM8(adr,val8); break; \
case 0x07 : GEN_CMP_rm8_imm8; break; \
}; \

#define GEN_GR1_rm16_imm16 \
DECODRM; \
GEN_GETRM16(adr,val16); \
switch(regop) \
{ \
case 0x00 : GEN_ADD_rm16_imm16; GEN_PUTRM16(adr,val16); break; \
case 0x01 : GEN_OR_rm16_imm16;  GEN_PUTRM16(adr,val16); break; \
case 0x02 : GEN_ADC_rm16_imm16; GEN_PUTRM16(adr,val16); break; \
case 0x03 : GEN_SBB_rm16_imm16; GEN_PUTRM16(adr,val16); break; \
case 0x04 : GEN_AND_rm16_imm16; GEN_PUTRM16(adr,val16); break; \
case 0x05 : GEN_SUB_rm16_imm16; GEN_PUTRM16(adr,val16); break; \
case 0x06 : GEN_XOR_rm16_imm16; GEN_PUTRM16(adr,val16); break; \
case 0x07 : GEN_CMP_rm16_imm16; break; \
}; \

#define GEN_GR1_rm16_rel8 \
DECODRM; \
GEN_GETRM16(adr,val16); \
switch(regop) \
{ \
case 0x00 : GEN_ADD_rm16_rel8; GEN_PUTRM16(adr,val16); break; \
case 0x01 : GEN_OR_rm16_rel8;  GEN_PUTRM16(adr,val16); break; \
case 0x02 : GEN_ADC_rm16_rel8; GEN_PUTRM16(adr,val16); break; \
case 0x03 : GEN_SBB_rm16_rel8; GEN_PUTRM16(adr,val16); break; \
case 0x04 : GEN_AND_rm16_rel8; GEN_PUTRM16(adr,val16); break; \
case 0x05 : GEN_SUB_rm16_rel8; GEN_PUTRM16(adr,val16); break; \
case 0x06 : GEN_XOR_rm16_rel8; GEN_PUTRM16(adr,val16); break; \
case 0x07 : GEN_CMP_rm16_rel8; break; \
}; \

#define GEN_GR2_rm8_1 \
FETCH(modrm); REGOP; RM; \
switch(regop) \
{ \
case 0x00 : GEN_ROL_rm8_1(rm); break; \
case 0x01 : GEN_ROR_rm8_1(rm); break; \
case 0x02 : GEN_RCL_rm8_1(rm); break; \
case 0x03 : GEN_RCR_rm8_1(rm); break; \
case 0x04 : GEN_SHL_rm8_1(rm); break; \
case 0x05 : GEN_SHR_rm8_1(rm); break; \
case 0x07 : GEN_SAR_rm8_1(rm); break; \
}; \

#define GEN_GR3_rm8 \
DECODRM; \
GEN_GETRM8(adr,val8); \
switch(regop) \
{ \
case 0x00 : GEN_TEST_rm8_imm8(val8); break; \
case 0x02 : GEN_NOT(val8); GEN_PUTRM8(adr,val8); break; \
case 0x03 : GEN_NEG8(val8); GEN_PUTRM8(adr,val8); IP+=1; break; \
case 0x04 : GEN_MUL_rm8(val8); break; \
case 0x05 : GEN_IMUL_rm8(val8); break; \
case 0x06 : GEN_DIV_rm8(val8); break; \
case 0x07 : GEN_IDIV_rm8(val8); break; \
}; \

#define GEN_GR3_rm16 \
DECODRM; \
GEN_GETRM16(adr,val16); \
switch(regop) \
{ \
case 0x00 : GEN_TEST_rm16_imm16(val16); break; \
case 0x02 : GEN_NOT(val16); GEN_PUTRM16(adr,val16); break; \
case 0x03 : GEN_NEG16(val16); GEN_PUTRM16(adr,val16); IP+=1; break; \
case 0x04 : GEN_MUL_rm16(val16); break; \
case 0x05 : GEN_IMUL_rm16(val16); break; \
case 0x06 : GEN_DIV_rm16(val16); break; \
case 0x07 : GEN_IDIV_rm16(val16); break; \
}; \

#define GEN_GR4_rm8 \
DECODRM; \
GEN_GETRM8(adr,val8); \
switch(regop) \
{ \
case 0x00 : GEN_INC8(val8); GEN_PUTRM8(adr,val8); break; \
case 0x01 : GEN_DEC8(val8); GEN_PUTRM8(adr,val8); break; \
}; \
IP+=1; \
